﻿//////////////////////////////////////////////////////////////////////////
// Copyright (c) 2011 Dynastream Innovations Inc.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
// WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
//
//////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Forms;
using System.Web;
using ANT_Managed_Library;
using System.Reflection;

namespace AutoSharedDemoApplication
{
    //Delegates for thread-safe calls
    delegate void dPrint(Control control, String msg);
    delegate void dEnabled(Control control);

    public partial class MainForm : Form
    {
        //Path to store log files in (if turned on). This puts them into the same directory as the executable
        String logDirectory = System.IO.Path.GetDirectoryName(Application.ExecutablePath);

        #region Constants
        /// <summary>
        /// Holds the value of the maximum ushort value
        /// </summary>
        private const ushort USHORT_MAX = 0xFFFF;

        /// <summary>
        /// Inital handshake "Acquire Address" message shared address
        /// </summary>
        private const ushort ASC_ACQUIRE_ADDRESS = 0xFFFF;
        public enum AcquireAddress : byte
        {
            ASC_ACQUIRE_ADDRESS0 = 0xFF,
            ASC_ACQUIRE_ADDRESS1 = 0xFF,
        }

        /// <summary>
        /// Handshake "Confirm Address" message shared address
        /// </summary>
        private const ushort ASC_CONFIRM_ADDRESS = 0xFFFE;
        public enum ConfirmAddress : byte
        {
            ASC_CONFIRM_ADDRESS0 = 0xFE,
            ASC_CONFIRM_ADDRESS1,
        }

        /// <summary>
        /// Shared channel message ID's
        /// </summary>
        public enum SharedChannelCommands : byte
        {
            SC_COMMAND_SHARED_CHANNEL_INFO = 0xFA,  //Channel Information (0xFA)
            SC_COMMAND_NO_SLOTS_AVAILABLE,          //No Address Available (0xFB)
            SC_COMMAND_CONFIRM_ACQUIRED,            //Confirm Acquire (0xFC)
            SC_COMMAND_REQUEST_TO_ACQUIRE,          //Request Address (0xFD)
            SC_COMMAND_BUSY_ACQUIRING,              //Busy Acquiring (0xFE)
            SC_COMMAND_SLOT_AVAILABLE,              //Address Available (0xFF)
        }

        /// <summary>
        /// Master Auto-shared Channel States
        /// </summary>
        public enum MacStates : byte
        {
            MAC_STATE_WAITING = 1,          //Waiting for a slave device to request address
            MAC_STATE_ACQUIRING,            //Acquiring slave device's unique ID, and giving it a shared channel address
            MAC_STATE_FULL,                 //No addresses available 
            MAC_SLOT_USED = 0x80,           //Address is used
            MAC_SLOT_WATCHDOG_MASK = 0x7F,  
            MAC_CONFIRM_RETIRES = 3,
            MAC_NO_CONNECTION_TIMEOUT = 5,
            MAC_ALIVE_CHECK_RETIRES = 5,
        }

        /// <summary>
        /// ANT device baud rates
        /// </summary>
        private enum BaudRates : ushort
        {
            AP1 = 50000,
            AP2AT3 = 57600,
        }

        #endregion

        #region Variables
        private ushort usMacAddress;                            //Holds address of the particular slave that the master is communicating with
        private ushort usMacAddressLimit;                       //Holds the number of slave devices for this particular application
        private ushort usDecimatedChannelPeriod;                //Holds the suggested channel period
        private byte ucDataTimeout;                             //Holds suggested timeout time
        private byte[] aucMacChannelStatus = new byte[255];     //Array holds status of available channel addresses 
        private ulong[] aulMacChannelID = new ulong[255];       //Array holds channel slave ID's
        private ushort usMacFreeSlotAddress;                    //Address of the next free slot on channel
        private ulong ulMacUniqueID;                            //Slave device unique ID
        private byte ucMacState;                                //Holds the state of the master
        private ushort usMacWatchdog;                           //Watchdog counter

        /// <summary>
        /// Holds the number of free slots
        /// </summary>
        private byte ucSlotsFree;

        bool bAcknowledged;
        byte ucResetMask = 0;

        bool bGetChannelData = false;

        private byte[] aucNetworkKey = new byte[] { 0, 0, 0, 0, 0, 0, 0, 0 };

        private byte ucUSBport = 0;

        #endregion

        /// <summary>
        /// Transmission buffer; information the Master will send to the slave devices
        /// </summary>
        private byte[] aucTxBuf = new byte[8];

        /// <summary>
        /// Array of slave device windows
        /// </summary>
        SlaveForm[] devices = new SlaveForm[8];

        #region ANT Device variables
        private static ANT_Device masterDevice;
        private static ANT_ReferenceLibrary.ChannelType masterChannelType = ANT_ReferenceLibrary.ChannelType.ADV_Shared_0x20 | ANT_ReferenceLibrary.ChannelType.BASE_Master_Transmit_0x10;
        private static ANT_Channel channel0;
        #endregion




        /// <summary>
        /// Constructor
        /// </summary>
        public MainForm()
        {
            InitializeComponent();

            //Display the version number on the upper left hand corner of the application form
            this.Text += "-v." + Assembly.GetExecutingAssembly().GetName().Version.Major + "." + Assembly.GetExecutingAssembly().GetName().Version.Minor;
            //Locks the form size to the current designer's size
            this.MinimumSize = this.Size;
            this.MaximumSize = this.Size;
        }

        /// <summary>
        /// Called when the Application Loads
        /// </summary>
        /// <param name="sender">Object calling event</param>
        /// <param name="e">event arguments</param>
        private void MainForm_Load(object sender, EventArgs e)
        {
            //Create 8 instances of SlaveForm 
            for (int i = 0; i < 8; i++)
            {
                devices[i] = new SlaveForm();
                devices[i].TopLevel = false;
                devices[i].Visible = true;
                flowLayoutPanel1.Controls.Add(devices[i]);

            }

            USBComboBoxPopulate();  //populate the USB combo box 
            ucResetMask = 255;      //set the reset mask to 255
            ResetForm();            //Reset the form
            NotConnected();         //Change control visibility/enabling 
            EnableLogging();        //Enable Debug logs, comment to turn off
        }

        #region Thread-safe calling methods
        /// <summary>
        /// Thread safe disable control method
        /// </summary>
        /// <param name="control">Control to be disabled</param>
        private void DisableControl(Control control)
        {
            if (control.InvokeRequired)
            {
                dEnabled d = new dEnabled(DisableControl);
                Invoke(d, new object[] { control });
            }
            else
            {
                control.Enabled = false;
            }
        }
        /// <summary>
        /// Thread safe enable control method
        /// </summary>
        /// <param name="control">Control to be enabled</param>
        private void EnableControl(Control control)
        {
            if (control.InvokeRequired)
            {
                dEnabled d = new dEnabled(EnableControl);
                Invoke(d, new object[]{control});
            }
            else
            {
                control.Enabled = true;
            }

        }
        /// <summary>
        /// Thread safe call to change control text
        /// </summary>
        /// <param name="control">Control to have text changed</param>
        /// <param name="msg">Text to be displayed on control</param>
        private void Print(Control control, String msg)
        {
            if (control.InvokeRequired)
            {
                dPrint d = new dPrint(Print);
                Invoke(d, new object[] { control, msg });
            }
            else
            {
                control.Text = msg;
            }
        }
        #endregion

        #region Reset Application/Devices
        /// <summary>
        /// Resets the application 
        /// </summary>
        private void ResetForm()
        {
            //Increment the reset mask
            ucResetMask++;

            //Reset and initialize each SlaveForm in the array
            for (int i = 0; i < 8; i++)
                ResetDevice(devices[i], i);

            usMacAddress = 0; //set the MacAddress to 0

            //Set each element in the MacChannelStatus array to 0x00
            foreach (byte i in aucMacChannelStatus)
                aucMacChannelStatus[i] = 0x00;

            //Set the first free slot to 1
            usMacFreeSlotAddress = 1;
            //Set the mac address limit to 7
            usMacAddressLimit = 8;
            Print(masterStatusTextBox, "");
            //The current state is waiting (0x01)
            ucMacState = (byte)(MacStates.MAC_STATE_WAITING);

            usDecimatedChannelPeriod = 4096; //8 Hz
            ucDataTimeout = 4; // 2*4 = 8s

            //Nothing has been Acknowledged yet
            bAcknowledged = false;
        }
        /// <summary>
        /// Resets the slave device frames
        /// </summary>
        /// <param name="device">Slave device fram to be reset</param>
        /// <param name="index">Index of slave device in array</param>
        private void ResetDevice(SlaveForm device, int index)
        {
            //Display the device number on the frame
            Print(device.DeviceNumber, ("Device " + (index + 1).ToString()));
            //Disable the device fram 
            DisableControl(device);
            //Set the status to inactive
            Print(device.Status, "Inactive");
            //Clear device ID
            Print(device.ID, "");
            //Clear sensor type
            Print(device.SensorType, "");
            //Clear the data received 
            Print(device.dataTextBox, "");
            //Print the device's shared address
            Print(device.sharedAddress, "[" + ((byte)((index + 1) % 256)).ToString("X2") + "][" + ((index + 1) / 256 + ucResetMask).ToString("X2") + "]");
            //Reset Data displayed
            Print(device.Data, "[AA][AA][AA][AA][AA][AA][AA][AA]");
            Print(device.typeTextBox, "");
        }

        #endregion

        /// <summary>
        /// Populate the USB combo box with the number of ports available
        /// </summary>
        private void USBComboBoxPopulate()
        {
            string[] astrAssignUSB = new string[] { "0", "1", "2", "3" };
            USBComboBox.Items.Clear();
            USBComboBox.Items.AddRange(astrAssignUSB);
        }

        /// <summary>
        /// Handle an index change in the USB combo-box
        /// </summary>
        /// <param name="sender">Object raising event</param>
        /// <param name="e">Event arguments</param>
        private void USBComboBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            ucUSBport = (byte)USBComboBox.SelectedIndex;
        }

        #region Button Click Event Handlers
        private void connectButton_Click(object sender, EventArgs e)
        {
            //Check if the button text is Connect Device or Disconnect Device
            if (connectButton.Text.Equals("Connect Device"))
            {
                try
                {
                    //assign the ANT usb connected to the compuer as the master device
                    masterDevice = new ANT_Device();
                    //put the application in "connected" state
                    Connected();
                    //Get channel 0 on the master device
                    channel0 = masterDevice.getChannel(0);
                    //Assign the channel response to the channel
                    channel0.channelResponse += new ANT_Channel.ChannelResponseHandler(ChannelResponse);
                    //Assign the device response to the device
                    masterDevice.deviceResponse += new ANT_Device.DeviceResponseHandler(DeviceResponse);
                }
                catch (ANT_Exception)
                {
                    //If no ANT device was detected, show an error box
                    MessageBox.Show("No Device Detected");
                }

            }
            else
            {
                //If the Disconnect button was clicked:
                //Reset the device
                masterDevice.ResetSystem();
                System.Threading.Thread.Sleep(500);
                //Destroy the instance of the ANT device that was created
                ANT_Device.shutdownDeviceInstance(ref masterDevice);
                //Put the application in "Not connected" state
                NotConnected();
            }
        }
        private void resetButton_Click(object sender, EventArgs e)
        {
            //Set the slots free variable to 8 
            ucSlotsFree = 8;
            //Reset the master device
            masterDevice.ResetSystem();
            System.Threading.Thread.Sleep(500);

            //Reset the application
            ResetForm();
            //Try setting the network key
            SetNetworkKey();
            //Initiate ANT configuration by setting the network key then waiting for the device response
            masterDevice.setNetworkKey(0, aucNetworkKey);
        }
        private void openButton_Click(object sender, EventArgs e)
        {
            //Check if the button text is "Open Channel" or "Close Channel"
            if (openButton.Text.Equals("Open Channel"))
            {
                try
                {
                    //try opening the channel
                    resetButton_Click(sender, e);
                    //Change the button text to "Close Channel"
                    Print(openButton, "Close Channel");
                    ConnectedAndOpen();
                }
                catch
                {
                    //Catch any exceptions that occur in the open channel operation
                    return;
                }
            }
            else
            {
                //Close the channel
                channel0.closeChannel();
                //Unassign the channel
                channel0.unassignChannel();
                //Change button text to "Open Channel"
                Print(openButton, "Open Channel");
                Connected();
            }
        }
        private void changeRFButton_Click(object sender, EventArgs e)
        {
            //Set channel 0's frequency to the value in the numeric up and down 
            channel0.setChannelFreq((byte)RFNumericUpDown.Value);
        }
        private void assignButton_Click(object sender, EventArgs e)
        {
            //Set the address to assign to the device number indicated in the numeric up and down
            ushort usAddressToAssign = (ushort)assignNumericUpDown.Value;

            //Mask in that the slot is used 
            aucMacChannelStatus[usAddressToAssign - 1] |= (byte)MacStates.MAC_SLOT_USED;
            //Set the ID to manual to indicate that the device was assigned manually
            Print(devices[usAddressToAssign - 1].ID, "Manual");
            //Set the Status to active
            Print(devices[usAddressToAssign - 1].Status, "Active");
            //Enable the device frame that was manually assigned
            EnableControl(devices[usAddressToAssign - 1]);
            //There is one less slot free
            ucSlotsFree--;

            if (usMacFreeSlotAddress == usAddressToAssign)
            {
                //Set the free slot address to 0
                usMacFreeSlotAddress = 0;
                //If the free slot addres is less than the address limit, and the device frame at the free slot address is used, increment the value of free slot addres
                while ((usMacFreeSlotAddress < usMacAddressLimit) && ((aucMacChannelStatus[usMacFreeSlotAddress] & (byte)MacStates.MAC_SLOT_USED) != 0))
                    usMacFreeSlotAddress++;

                //If the free slot address is less than the address limit, increment the free slot address and throw the master into waiting state
                if (usMacFreeSlotAddress < usMacAddressLimit)
                {
                    usMacFreeSlotAddress++;
                    ucMacState = (byte)MacStates.MAC_STATE_WAITING;
                }
                else
                {
                    //If the free slot address is not less than the address limit, set the free slot address to 0, and  indicate that there are no more available slots
                    usMacFreeSlotAddress = 0;
                    ucMacState = (byte)MacStates.MAC_STATE_FULL;
                }
            }
        }
        #endregion

        #region Application States
        private void NotConnected()
        {
            //This state is used when there is no device connected to the application 
            Print(connectButton, "Connect Device");
            Print(openButton, "Open Channel");
            DisableControl(DeviceControls);
            DisableControl(ChannelSettings);
            DisableControl(DemoControls);
        } 
        private void Connected()
        {
            //This state is used when a device is connected, but the channel is not open
            Print(connectButton, "Disconnect Device");
            DisableControl(USBComboBox);
            DisableControl(DemoControls);
            EnableControl(ChannelSettings);
            EnableControl(DeviceControls);
            Print(masterStatusTextBox, "Device Connected: " + ucSlotsFree.ToString() + " Slots free");
        }
        private void ConnectedAndOpen()
        {
            //This state is used when a device is connected and the channel is open
            Print(connectButton, "Disconnect Device");
            DisableControl(USBComboBox);
            EnableControl(DemoControls);
            DisableControl(ChannelSettings);
            Print(openButton, "Close Channel");
        }
        #endregion

        /// <summary>
        /// Handles the Channel response 
        /// </summary>
        /// <param name="response">Channel response message</param>
        private void ChannelResponse(ANT_Response response)
        {
            //Check Message ID of the response 
            switch ((ANT_ReferenceLibrary.ANTMessageID)response.responseID)
            {
                case ANT_ReferenceLibrary.ANTMessageID.RESPONSE_EVENT_0x40:
                    {
                        //Get the channel Event code of the message received
                        switch (response.getChannelEventCode())
                        {
                            case ANT_ReferenceLibrary.ANTEventID.EVENT_TRANSFER_TX_COMPLETED_0x05:
                                {
                                    //Check if the slot is used
                                    if ((aucMacChannelStatus[usMacAddress - 1] & (byte)(MacStates.MAC_SLOT_USED)) != 0)
                                        //Clear the channel watchdog
                                        aucMacChannelStatus[usMacAddress - 1] = (byte)(MacStates.MAC_SLOT_USED);
                                    Print(devices[usMacAddress - 1].sendButton, "Send");
                                    goto case ANT_ReferenceLibrary.ANTEventID.EVENT_TX_0x03; //Fall through to the next case
                                }
                            case ANT_ReferenceLibrary.ANTEventID.EVENT_TX_0x03:
                                {
                                    usMacAddress--; //Advance to the next slot address

                                    if (usMacAddress == USHORT_MAX) //Check if the decrement yields 65535
                                        usMacAddress = usMacAddressLimit; //Rollover to address limit 

                                    //Check that the slot address is not zero, and check if the slot is used
                                    if ((usMacAddress != 0) && ((aucMacChannelStatus[usMacAddress - 1] & (byte)(MacStates.MAC_SLOT_USED)) != 0))
                                    {
                                        //If this slot is in use:

                                        //Write MacAddress into first two bytes of Tx Buffer
                                        aucTxBuf[0] = BitConverter.GetBytes(usMacAddress)[0];
                                        aucTxBuf[1] = BitConverter.GetBytes(usMacAddress)[1];
                                        //Mask in the reset mask
                                        aucTxBuf[1] |= ucResetMask;
                                        //Write default shared info message
                                        aucTxBuf[2] = (byte)(SharedChannelCommands.SC_COMMAND_SHARED_CHANNEL_INFO);

                                        //Give suggested data timeout to slave in message
                                        aucTxBuf[5] = ucDataTimeout;
                                        //Give suggested channel period decimation value in message 
                                        aucTxBuf[6] = BitConverter.GetBytes(usDecimatedChannelPeriod)[0];
                                        aucTxBuf[7] = BitConverter.GetBytes(usDecimatedChannelPeriod)[1];

                                        //Overwrite the default info message with any data from the device
                                        //frame that needs to be passed to the devices
                                        byte[] temp = new byte[]{0, 0, 0, 0, 0, 0};
                                        Array.Copy(aucTxBuf, 2, temp, 0, 6); 

                                        //Check if there is any data to be passed to the slave
                                        bAcknowledged = devices[usMacAddress - 1].GetChannelData(ref temp, ref bGetChannelData);
                                        
                                        //If Data needs to be passed to the slave, write the data into the TxBuffer
                                        if (bGetChannelData)
                                        {
                                            Array.Copy(temp, 0, aucTxBuf, 2, 6);
                                            bGetChannelData = false;
                                        }

                                        if ((aucMacChannelStatus[usMacAddress - 1] & (byte)(MacStates.MAC_SLOT_WATCHDOG_MASK)) != 0)
                                        {
                                            //If the watchdog counter runs out
                                            if (((--aucMacChannelStatus[usMacAddress - 1]) & (byte)(MacStates.MAC_SLOT_WATCHDOG_MASK)) == 0)
                                            {
                                                //Free up the channel because the channel alive watchdog ran out
                                                aucMacChannelStatus[usMacAddress - 1] = 0;

                                                //If we were full before
                                                if (ucMacState == (byte)(MacStates.MAC_STATE_FULL))
                                                {
                                                    ucMacState = (byte)(MacStates.MAC_STATE_WAITING);
                                                    usMacFreeSlotAddress = usMacAddress;
                                                }
                                                DisableControl(devices[usMacAddress - 1]);
                                                Print(devices[usMacAddress - 1].Status, "Inactive");
                                                Print(devices[usMacAddress - 1].sendButton, "Send");
                                            }
                                        }
                                        else
                                        {
                                            if (bAcknowledged)
                                                //Set the watchdog counter
                                                aucMacChannelStatus[usMacAddress - 1] = ((byte)((MacStates.MAC_SLOT_USED) | (MacStates.MAC_ALIVE_CHECK_RETIRES)));
                                        }

                                    }
                                    else //If the slot is free
                                    {
                                        //If this is a slot being acquired and this address is the free slot address 
                                        if ((ucMacState == (byte)(MacStates.MAC_STATE_ACQUIRING)) && (usMacFreeSlotAddress == usMacAddress))
                                        {
                                            //Makesure the watchdog is non-zero
                                            if ((usMacWatchdog--) != 0)
                                            {
                                                //Write the address into the first 2 bytes of the Tx Buffer
                                                aucTxBuf[0] = BitConverter.GetBytes(usMacAddress)[0];
                                                aucTxBuf[1] = BitConverter.GetBytes(usMacAddress)[1];
                                                //Mask in the reset mask
                                                aucTxBuf[1] |= ucResetMask;
                                            }
                                            else //This is not a slot being acquired
                                            {
                                                //Write 0xFFFF as the MAC Address, broadcast to all devices
                                                aucTxBuf[0] = (byte)(AcquireAddress.ASC_ACQUIRE_ADDRESS0);
                                                aucTxBuf[1] = (byte)(AcquireAddress.ASC_ACQUIRE_ADDRESS1);
                                                //Exit acquiring mode if we do not receive a confirmation after all of the retries  have passed
                                                ucMacState = (byte)(MacStates.MAC_STATE_WAITING);
                                            }
                                        }
                                        else //This slot is free
                                        {
                                            //Write 0xFFFF as the MAC Address
                                            aucTxBuf[0] = (byte)(AcquireAddress.ASC_ACQUIRE_ADDRESS0);
                                            aucTxBuf[1] = (byte)(AcquireAddress.ASC_ACQUIRE_ADDRESS1);
                                        }

                                        //If the master is acquiring
                                        if (ucMacState == (byte)(MacStates.MAC_STATE_ACQUIRING))
                                        {
                                            //Write 0xFFFE as the Shared address
                                            aucTxBuf[0] = (byte)(ConfirmAddress.ASC_CONFIRM_ADDRESS0);
                                            aucTxBuf[1] = (byte)(ConfirmAddress.ASC_CONFIRM_ADDRESS1);
                                            aucTxBuf[2] = (byte)(SharedChannelCommands.SC_COMMAND_BUSY_ACQUIRING);
                                            //Write slave's unique ID into last 4 bytes of Tx Buffer 
                                            //(Unique ID received from slave in 0xFD message)
                                            aucTxBuf[4] = BitConverter.GetBytes(ulMacUniqueID)[0];
                                            aucTxBuf[5] = BitConverter.GetBytes(ulMacUniqueID)[1];
                                            aucTxBuf[6] = BitConverter.GetBytes(ulMacUniqueID)[2];
                                            aucTxBuf[7] = BitConverter.GetBytes(ulMacUniqueID)[3];
                                        }
                                        else
                                        {
                                            //If there are no free slots
                                            if (usMacFreeSlotAddress == 0)
                                            {
                                                //Set the default message as no slots available
                                                aucTxBuf[2] = (byte)(SharedChannelCommands.SC_COMMAND_NO_SLOTS_AVAILABLE);
                                                Print(masterStatusTextBox, "No Slots Free");
                                            }
                                            else//There are free slots
                                            {
                                                //Set the default message to slots available
                                                aucTxBuf[2] = (byte)(SharedChannelCommands.SC_COMMAND_SLOT_AVAILABLE);
                                                Print(masterStatusTextBox, "Device Connected: " + ucSlotsFree.ToString() + " Slots free");
                                            }
                                            //Set data timeout to be 8 seconds
                                            aucTxBuf[5] = ucDataTimeout;
                                            //Write Free slot address into Tx Buffer
                                            aucTxBuf[6] = BitConverter.GetBytes(usMacFreeSlotAddress)[0];
                                            aucTxBuf[7] = BitConverter.GetBytes(usMacFreeSlotAddress)[1];
                                            //Mask in reset mask
                                            aucTxBuf[7] |= ucResetMask;
                                        }
                                    }

                                    //If an acknoledgement is required
                                    if (bAcknowledged)
                                    {
                                        //Send txBuffer as acknowledged data 
                                        bAcknowledged = false;
                                        channel0.sendAcknowledgedData(aucTxBuf);
                                    }
                                    else//No acknowledgement is required
                                    {
                                        //Send txBuff as broadcast data
                                        channel0.sendBroadcastData(aucTxBuf);
                                    }
                                    break;
                                }

                            case ANT_ReferenceLibrary.ANTEventID.EVENT_TRANSFER_TX_FAILED_0x06:
                                {
                                    //set the event ID as EVENT_TX
                                    response.messageContents[2] = 0x03;
                                    //re-call the Channel response handler
                                    ChannelResponse(response);
                                    break;
                                }
                        }

                        break;
                    }
                case ANT_ReferenceLibrary.ANTMessageID.ACKNOWLEDGED_DATA_0x4F:
                    {
                        //create a temporary array
                        char[] acTemp = new char[9];

                        //If acknowledged data is required:
                        if (bAcknowledged)
                        {
                            bAcknowledged = false;
                            channel0.sendAcknowledgedData(aucTxBuf);
                        }

                        //Set usMasgMacAddress to the address of the slave that the acknowledged data was received from
                        ushort usMesgMacAddres = BitConverter.ToUInt16(response.messageContents, 1);
                        //Set the MesgUniqueID to the unique ID received from the slave
                        ulong ulMesgUniqueID = BitConverter.ToUInt32(response.messageContents, 5);

                        //Check the Message ID of the acknowledge data
                        switch ((byte)response.messageContents[3])
                        {
                            case (byte)SharedChannelCommands.SC_COMMAND_REQUEST_TO_ACQUIRE:
                                {
                                    //If the slave is requesting to acquire and the master is in the wait state
                                    if ((usMesgMacAddres == ASC_ACQUIRE_ADDRESS) && (ucMacState == (byte)MacStates.MAC_STATE_WAITING))
                                    {
                                        //Start acquiring the slave device's unique ID 
                                        usMacWatchdog = (byte)MacStates.MAC_CONFIRM_RETIRES;
                                        ulMacUniqueID = ulMesgUniqueID;
                                        //Set the state as acquiring
                                        ucMacState = (byte)MacStates.MAC_STATE_ACQUIRING;
                                        Print(masterStatusTextBox, "Acquiring...");

                                        Print(devices[usMacFreeSlotAddress - 1].Status, "Acquiring...");
                                        Print(devices[usMacFreeSlotAddress - 1].ID, ulMacUniqueID.ToString());
                                        EnableControl(devices[usMacFreeSlotAddress - 1]);
                                    }
                                    break;
                                }
                            case (byte)SharedChannelCommands.SC_COMMAND_CONFIRM_ACQUIRED:
                                {
                                    //If the slave is requesting a confirmation of acquire
                                    if ((ulMesgUniqueID == ulMacUniqueID) && (ucMacState == (byte)MacStates.MAC_STATE_ACQUIRING))
                                    {
                                        //Mark the slot as assigned
                                        aucMacChannelStatus[usMacFreeSlotAddress - 1] |= (byte)MacStates.MAC_SLOT_USED;
                                        //Hold the unique ID in the Channel ID array
                                        aulMacChannelID[usMacFreeSlotAddress - 1] = ulMacUniqueID;

                                        Print(devices[usMacFreeSlotAddress - 1].Status, "Active");
                                        EnableControl(devices[usMacFreeSlotAddress - 1]);

                                        //Set free slot address to 0
                                        usMacFreeSlotAddress = 0;
                                        //Check that the slot address is less than the address limit, and check if the slot is used
                                        while ((usMacFreeSlotAddress < usMacAddressLimit) && ((aucMacChannelStatus[usMacFreeSlotAddress] & (byte)MacStates.MAC_SLOT_USED) != 0))
                                            usMacFreeSlotAddress++;

                                        //IF the free slot address is less than the limit
                                        if (usMacFreeSlotAddress < usMacAddressLimit)
                                        {
                                            //Index is offset by 1
                                            usMacFreeSlotAddress++;
                                            //Decrement the slots free variable
                                            ucSlotsFree--;
                                            //Put the master in waiting state
                                            ucMacState = (byte)MacStates.MAC_STATE_WAITING;
                                        }
                                        else
                                        {
                                            //All of the slots are used
                                            usMacFreeSlotAddress = 0;
                                            //put master in full state
                                            ucMacState = (byte)MacStates.MAC_STATE_FULL;
                                        }
                                    }
                                    break;
                                }
                            default:
                                {
                                    byte[] temp = response.getDataPayload();
                                    DisplayData(temp);
                                    break;
                                }
                        }

                        break;
                    }

                case ANT_ReferenceLibrary.ANTMessageID.BROADCAST_DATA_0x4E:
                    {
                        byte[] temp = response.getDataPayload();  
                        DisplayData(temp);
                        break;
                    }
            }
        }

        /// <summary>
        /// Handles the Device response
        /// </summary>
        /// <param name="response">Response from ANT</param>
        private void DeviceResponse(ANT_Response response)
        {
            try
            {
                switch (response.getMessageID())
                {
                    case ANT_ReferenceLibrary.ANTMessageID.UNASSIGN_CHANNEL_0x41:
                        break;
                    case ANT_ReferenceLibrary.ANTMessageID.ASSIGN_CHANNEL_0x42:
                        {
                            if ((ANT_ReferenceLibrary.ANTEventID)response.messageContents[2] == ANT_ReferenceLibrary.ANTEventID.RESPONSE_NO_ERROR_0x00)
                            {
                                //If the Assign Channel message was detected, and ANT responded with RESPONSE_NO_ERROR
                                //(i.e. the channel was successfully assigned as master shared) set the Channel ID
                                //After the channel is successfully assigned, set the channel ID
                                channel0.setChannelID((byte)deviceNoUpDown.Value, false, (byte)deviceTypeUpDown.Value, (byte)transTypeUpDown.Value);
                            }
                            break;
                        }
                    case ANT_ReferenceLibrary.ANTMessageID.NETWORK_KEY_0x46:
                        {
                            if ((ANT_ReferenceLibrary.ANTEventID)response.messageContents[2] == ANT_ReferenceLibrary.ANTEventID.RESPONSE_NO_ERROR_0x00)
                            {
                                //If the Network Key message was detected, and ANT responded with RESPONSE_NO_ERROR 
                                //(i.e. the Netwok Key was successfully set) Assign the Channel as a master shared channel
                                channel0.assignChannel(masterChannelType, 0);
                            }
                            break;
                        }
                    case ANT_ReferenceLibrary.ANTMessageID.CHANNEL_ID_0x51:
                        {
                            if ((ANT_ReferenceLibrary.ANTEventID)response.messageContents[2] == ANT_ReferenceLibrary.ANTEventID.RESPONSE_NO_ERROR_0x00)
                            {
                                //If the Channel ID Message was detected, and ANT responded with RESPONSE_NO_ERROR
                                //(i.e. the Channel ID was successfully set) Set the Channel Period to 8 Hz
                                channel0.setChannelPeriod(4096);
                            }
                            break;
                        }
                    case ANT_ReferenceLibrary.ANTMessageID.CHANNEL_MESG_PERIOD_0x43:
                        {
                            if ((ANT_ReferenceLibrary.ANTEventID)response.messageContents[2] == ANT_ReferenceLibrary.ANTEventID.RESPONSE_NO_ERROR_0x00)
                            {
                                //If the Channel Message Period message was detected, and ANT responded with RESPONSE_NO_ERROR
                                //(i.e. the message period was successfully set) set the Channel Frequency 
                                channel0.setChannelFreq((byte)radioNumericUpDown.Value);
                            }
                            break;
                        }
                    case ANT_ReferenceLibrary.ANTMessageID.CHANNEL_RADIO_FREQ_0x45:
                        {
                            if ((ANT_ReferenceLibrary.ANTEventID)response.messageContents[2] == ANT_ReferenceLibrary.ANTEventID.RESPONSE_NO_ERROR_0x00)
                            {
                                //If the Channel Radio Frequency message was detected, and ANT responded with RESPONSE_NO_ERROR
                                //(i.e. the radio frequency was successfully set) Open the Channel
                                channel0.openChannel();
                            }
                            break;
                        }

                    case ANT_ReferenceLibrary.ANTMessageID.CHANNEL_SEARCH_TIMEOUT_0x44:
                        break;
                    case ANT_ReferenceLibrary.ANTMessageID.OPEN_CHANNEL_0x4B:
                        break;
                    case ANT_ReferenceLibrary.ANTMessageID.CLOSE_CHANNEL_0x4C:
                        break;
                    default:
                        break;
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show("Device Response processing failed with exception: " + ex.Message);
            }
        }

        /// <summary>
        /// Display the Data received from a slave device 
        /// </summary>
        /// <param name="aucData">Data to decode and display</param>
        private void DisplayData(byte[] aucData)
        {
            //Set the shared address to the first two bytes of the data received from the slave device
            ushort usSharedAddress;
            usSharedAddress = (ushort)((aucData[1] << 8) + aucData[0]);

            //Check that the shared address is non-zero, not "Address aquire(0xFFFF)", and not "Confirm address(0xFFFE)"
            if ((usSharedAddress != 0) && (usSharedAddress != ASC_ACQUIRE_ADDRESS) && (usSharedAddress != ASC_CONFIRM_ADDRESS))
            {
                //mask off any of the reset mask bits
                ushort number = 0xFF00;
                usSharedAddress &= (ushort)(~(number));
                //Display the data in the slave's window
                devices[usSharedAddress - 1].DisplayData(aucData);
            }
        }

        /// <summary>
        /// Handle the data typed into the network key textbox
        /// Allow only hexadecimal numbers/letters to be input
        /// </summary>
        /// <param name="sender">Object calling event</param>
        /// <param name="e">Key event arguments</param>
        private void networkMaskedTextBox_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode > Keys.F && e.KeyCode <= Keys.Z && !(e.Control) && !(e.Alt))
                e.SuppressKeyPress = true; //Suppress any Keys that aren't 0-9 or A-F
        }

        /// <summary>
        /// Sets the network key from the data displayed in the network key textbox
        /// </summary>
        private void SetNetworkKey()
        {
            //Set a temporary string to the text in the network key textbox
            string s = networkMaskedTextBox.Text;
            string t = "";
            bool newbyte = false;
            byte temp = 0;
            int temp2 = 0;
            byte value = 0;

            //If the data is not input correctly, display an error message and throw an exception
            if (!networkMaskedTextBox.MaskFull)
            {
                MessageBox.Show("Network Key in incorrect format!", "Invalid Data", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                throw new Exception();
            }

            //Conver the network key string to an array of bytes
            for (int i = 0, k = 0; i < 23; i++)
            {
                temp = 0;
                t = s.Substring(i, 1);
                while (!t.Equals("-"))
                {
                    value = HexValue(t);
                    if (!newbyte)
                    {
                        temp2 = (int)value << 4;
                        newbyte = true;
                    }
                    else
                    {
                        temp = value;
                        temp += (byte)temp2;
                        newbyte = false;
                        aucNetworkKey[k] = temp;
                        k++;
                    }

                    if (i == 22)
                        return;
                    i++;
                    t = s.Substring(i, 1);
                }
            }
        }

        /// <summary>
        /// Return the value of a hexadecimal number in decimal
        /// </summary>
        /// <param name="t">Hexadecimal string</param>
        /// <returns>Value of the hexadecimal string in decimal</returns>
        private byte HexValue(string t)
        {
            byte value;
            if (t.Contains("A"))
                value = 10;
            else
                if (t.Contains("B"))
                    value = 11;
                else
                    if (t.Contains("C"))
                        value = 12;
                    else
                        if (t.Contains("D"))
                            value = 13;
                        else
                            if (t.Contains("E"))
                                value = 14;
                            else
                                if (t.Contains("F"))
                                    value = 15;
                                else
                                    value = byte.Parse(t);

            return value;

        }

        /// <summary>
        /// Handles the Form Closing Event
        /// </summary>
        /// <param name="sender">Object raising event</param>
        /// <param name="e">Form closing event arguments</param>
        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            //Check if the ANT device is connected when the form is closed
            if (connectButton.Text.Equals("Disconnect Device"))
            {
                channel0.closeChannel();
                channel0.unassignChannel();
                masterDevice.ResetSystem();
                System.Threading.Thread.Sleep(500);
            }
            ANT_Device.shutdownDeviceInstance(ref masterDevice);
        }

        /// <summary>
        /// Allow log files in the application
        /// </summary>
        private void EnableLogging()
        {
            //Enable log and make sure it was successful
            if (!ANT_Common.enableDebugLogs())
            {
                MessageBox.Show("Enabling logs failed. Communication will not be recorded", "Log File Failure");
                return;
            }
            //Catch any file errors that might exist
            try
            {
                //Create the log directory if it doesn't already exist
                if (!System.IO.Directory.Exists(logDirectory))
                    System.IO.Directory.CreateDirectory(logDirectory);
                //Set the library to save logs in the log directory 
                if (!ANT_Common.setDebugLogDirectory(logDirectory + "\\"))
                    throw new Exception("Could not set log file directory");
            }
            catch (Exception)
            {
                MessageBox.Show("Enabling log files failed, could not set log directory, ensure application is running with appropriate permission to creat files. Communication will not be recorded.", "Log File Failure");
                return;
            }
        }
    }
}
